/*******************************************************************************

WePhoto - Photo viewer for multi-touch devices
Copyright (C) 2010 Mihai Paslariu
Original Name MihPhoto
Adapted for WeTab 2011 by Peter Herzog www.pherzog.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include <QtGui>
#include <stdio.h>

#include "Config.h"
#include "ImageArea.h"

/*******************************************************************************
* CONSTRUCTOR / DESTRUCTOR
*******************************************************************************/

ImageArea::ImageArea(QWidget *parent)
    : QWidget(parent)
{
	setAttribute(Qt::WA_StaticContents);
	_current_index = 0;
	_drag_offset = 0;
	_rotation = _initial_rotation = 0.0;
	_initial_zoom = 0.0;

	_previous = _current = _next = NULL;
	_posx = _posy = 0;
	_previous_zoom = _current_zoom = _next_zoom = 1.0;
	_mouse_start_x = _mouse_start_y = 0;
	_mouse_drag = false;

	_initial_posx = _initial_posy = _initial_drag_offset = 0;
	_allow_zoom = _allow_pan = _allow_drag = false;
	_changing = false;
	_two_fingers = false;

	connect( &_timer, SIGNAL(timeout()), this, SLOT(onTimer()));
	_timer.start( 30 );

	_ui.addAction( TOUCH_ACTION_OPEN, "document-open.png" );
	_ui.addAction( TOUCH_ACTION_FULLSCREEN, "view-fullscreen.png" );
	_ui.addAction( TOUCH_ACTION_EXIT, "application-exit.png" );
	_ui.addSeparator();

	_ui.addAction( TOUCH_ACTION_GO_FIRST, "go-first.png" );
	_ui.addAction( TOUCH_ACTION_GO_PREVIOUS, "go-previous.png" );
	_ui.addAction( TOUCH_ACTION_GO_NEXT, "go-next.png" );
	_ui.addAction( TOUCH_ACTION_GO_LAST, "go-last.png" );
	_ui.addSeparator();

	_ui.addAction( TOUCH_ACTION_ZOOM_IN, "zoom-in.png" );
	_ui.addAction( TOUCH_ACTION_ZOOM_OUT, "zoom-out.png" );
	_ui.addAction( TOUCH_ACTION_ZOOM_FIT, "zoom-fit-best.png" );
	_ui.addAction( TOUCH_ACTION_ZOOM_ORIGINAL, "zoom-original.png" );
	_ui.addAction( TOUCH_ACTION_ROTATE_LEFT, "object-rotate-left.png" );
	_ui.addAction( TOUCH_ACTION_ROTATE_RIGHT, "object-rotate-right.png" );
	_ui.addSeparator();


	_zoom_in_icon.load(g_config.install_dir + "zoom-in.png" );
	_zoom_out_icon.load(g_config.install_dir + "zoom-out.png" );
	_zoom_indicator.load(g_config.install_dir + "position.png" );

	_load_thread.start();
	_show_ui = false;

	if ( g_config.multitouch )
	{
		setAttribute(Qt::WA_AcceptTouchEvents);
	}
	else
	{
		setMouseTracking(true);
	}
}

ImageArea::~ImageArea()
{
	_load_thread.stopThread();
	_load_thread.wait();
	delete _current;
	delete _next;
	delete _previous;
}

/*******************************************************************************
* PUBLIC METHODS AND SLOTS
*******************************************************************************/

void ImageArea::setFiles( QStringList files, QString current )
{
	_current_index = 0;
	_files = files;
	//for ( int i = 0; i < files.size(); i++ )
	//	printf("%s\n", files[i].toUtf8().data() );
	for ( int i = 0; i < files.size(); i++ )
	{
		if ( current == files[i] )
		{
			_current_index = i;
			break;
		}
	}
	_reloadAll();
}

void ImageArea::forward( void )
{
	if ( !_changing && _current_index < _files.size() - 1 && _load_thread.isIdle() )
	{
		_allow_drag = false;
		_current_zoom = _computeFitZoom( _current );
		_next_zoom = _computeFitZoom( _next );
		_posx = _posy = 0;
		_rotation = 0.0;
		_timer.start( 30 );

		_drag_offset = width();
		_moveForward();
		_changing = true;
	}
}

void ImageArea::back( void )
{
	if ( !_changing && _current_index > 0 && _load_thread.isIdle() )
	{
		_allow_drag = false;
		_current_zoom = _computeFitZoom( _current );
		_previous_zoom = _computeFitZoom( _previous );
		_posx = _posy = 0;
		_rotation = 0.0;
		_timer.start( 30 );

		_drag_offset = -width();
		_moveBack();
		_changing = true;
	}
}

void ImageArea::first( void )
{
	if ( !_changing && _current_index != 0 && _load_thread.isIdle() )
	{
		_rotation = 0.0;
		_allow_drag = false;
		_drag_offset = 0;
		_current_index = 0;
		_reloadAll();
	}
}

void ImageArea::last( void )
{
	if ( !_changing && _current_index != _files.count()-1 && _load_thread.isIdle() )
	{
		_rotation = 0.0;
		_allow_drag = false;
		_drag_offset = 0;
		_current_index = _files.count()-1;
		_reloadAll();
	}
}

void ImageArea::changeUI( void )
{
	_show_ui = !_show_ui;
	update();
}

void ImageArea::zoomIn( double speed )
{
	if ( _current == NULL || _changing )
		return;

	speed = ( speed > 0.0 ? speed : g_config.zoom_speed );
	_current_zoom /= speed;
	_limitZoom(_current_zoom);
	update();
}

void ImageArea::zoomOut( double speed )
{
	if ( _current == NULL || _changing )
		return;

	speed = ( speed > 0.0 ? speed : g_config.zoom_speed );
	_current_zoom *= speed;
	
	// limit zoom
	_limitZoom(_current_zoom);
	double fit_zoom = _computeFitZoomWithRotation(_current);
	double min_zoom = ( fit_zoom < 1.0 ? fit_zoom : 1.0 );
	if ( _current_zoom < min_zoom )
		_current_zoom = min_zoom;
	
	update();
}

void ImageArea::zoomToFit( void )
{
	if ( _current == NULL || _changing )
		return;

	_rotation = 0.0;
	_posx = _posy = 0;
	_current_zoom = _computeFitZoomWithRotation(_current);
	update();
}

void ImageArea::zoomToOriginalSize( void )
{
	if ( _current == NULL || _changing )
		return;

	_current_zoom = 1.0;
	_rotation = 0.0;
	_posx = _posy = 0;
	update();
}

void ImageArea::rotateLeft( void )
{
	if ( _current == NULL || _changing )
		return;

	double fz = _computeFitZoomWithRotation(_current);
	int r = (int)_rotation;
	r = r/90*90 + ( r % 90 < 45 ? 0 : 90 ) + 90;
	if ( r < 0 ) r = r + 360 * (abs(r)/360 + 1);
	if ( r >= 360 ) r = r - 360 * r/360;
	_rotation = (double)r;
	double zdif = fz - _current_zoom;
	if ( fabs(zdif) < 0.01 )
		_current_zoom = _computeFitZoomWithRotation(_current);
	update();
}

void ImageArea::rotateRight( void )
{
	if ( _current == NULL || _changing )
		return;

	double fz = _computeFitZoomWithRotation(_current);
	int r = (int)_rotation;
	r = r/90*90 + ( r % 90 < 45 ? 0 : 90 ) - 90;
	if ( r < 0 ) r = r + 360 * (abs(r)/360 + 1);
	if ( r >= 360 ) r = r - 360 * r/360;
	_rotation = (double)r;
	double zdif = fz - _current_zoom;
	if ( fabs(zdif) < 0.01 )
		_current_zoom = _computeFitZoomWithRotation(_current);
	update();
}

void ImageArea::moveView( int deltax, int deltay )
{
	if ( _current == NULL || _changing )
		return;

	_posx -= deltax;
	_posy -= deltay;
	_limitPan();
}

/*******************************************************************************
* EVENTS
*******************************************************************************/

void ImageArea::paintEvent(QPaintEvent *event)
{
	QPainter painter(this);
	const QRect rect = event->rect();
	painter.fillRect( rect, g_config.background );

	int cx = width() / 2;
	int cy = height() / 2;

	/*if ( _current )
	{
		int sw = (int)( (double)_current->width() * _current_zoom );
		int sh = (int)( (double)_current->height() * _current_zoom );
		int x = cx - sw/2 + _posx + _drag_offset;
		int y = cy - sh/2 + _posy;
		QRect r(x,y,sw,sh);
		painter.drawImage( r, *_current );
	}*/

	if ( _current )
	{
		QTransform tr = _getCurrentTransform();
		painter.setTransform(tr);
		painter.drawImage( 0,0, *_current );
		painter.resetTransform();
	} else {
		QString s1 = "", s2 = "";
		if ( _files.empty() )
		{
			s1 = tr("Kein Bild gefunden.");
			s2 = tr("Klicken sie auf den Schirm um das Menue zu zeigen!");
		} else {
			if ( _load_thread.isIdle() )
			{
				s1 = tr("Kann Bild nicht laden.");
				s2 = _files[_current_index];
			} else {
				s1 = tr("Lade.");
				s2 = tr("Bitte warten!");
			}
		}

		int w,h,x,y;
		QFontMetrics fm( painter.font() );
		painter.setPen(Qt::white);
		w = fm.width(s1);
		h = fm.height();
		x = ( width() - w ) / 2 + _drag_offset;
		y = height()/2 - h - 2;
		painter.drawText( x,y, s1 );
		w = fm.width(s2);
		x = ( width() - w ) / 2 + _drag_offset;
		y = height()/2 + 2;
		painter.drawText( x,y, s2 );

	}

	if ( _next && _drag_offset < 0 )
	{
		int sw = (int)( (double)_next->width() * _next_zoom );
		int sh = (int)( (double)_next->height() * _next_zoom );
		int x = cx - sw/2 + _posx + _drag_offset + width();
		int y = cy - sh/2 + _posy;
		QRect r(x,y,sw,sh);
		painter.drawImage( r, *_next );
	}

	if ( _previous && _drag_offset > 0 )
	{
		int sw = (int)( (double)_previous->width() * _previous_zoom );
		int sh = (int)( (double)_previous->height() * _previous_zoom );
		int x = cx - sw/2 + _posx + _drag_offset - width();
		int y = cy - sh/2 + _posy;
		QRect r(x,y,sw,sh);
		painter.drawImage( r, *_previous );
	}

	if ( g_config.show_file_name )
	{
		QString s;
		if ( _files.count() != 0 )
		{
			s.sprintf("[%d/%d] ", (_current_index+1), _files.count() );
			s += _files[_current_index];
		}


		int x = 10,y = 5;
		QFontMetrics fm( painter.font() );
		int w = fm.width(s);
		int h = fm.height();
		painter.setBrush(g_config.background);
		painter.setPen(Qt::transparent);
		painter.drawRect(0,0,w+x*2,h+y*2);
		painter.setPen(Qt::white);
		painter.drawText( x,y+h, s );
	}

	if ( _show_ui )
		_ui.draw( painter );
	if ( _show_ui && !g_config.multitouch )
	{
		int x,y;
		
		
		QRect rect(0,height() - _zoom_out_icon.height(),
			painter.window().width(),painter.window().height());
		painter.drawTiledPixmap( rect, *_ui.getBackground() );
		x = 0;
		y = height() - _zoom_out_icon.height();
		painter.drawImage( x,y, _zoom_out_icon );
		x = width() - _zoom_out_icon.width();
		y = height() - _zoom_in_icon.height();
		painter.drawImage( x,y, _zoom_in_icon );
		if ( _current != NULL )
		{
			double fit_zoom = _computeFitZoomWithRotation(_current);
			double min_zoom = ( fit_zoom < 1.0 ? fit_zoom : 1.0 );
			double max_zoom = g_config.max_zoom;
			double pos = ( _current_zoom - min_zoom ) / ( max_zoom - min_zoom );
			int w = width() - _zoom_in_icon.width() - _zoom_out_icon.width();
			x = _zoom_out_icon.width() + pos * w  - _zoom_indicator.width()/2;
		} else {
			x = width() / 2 - _zoom_indicator.width()/2;
		}
		y = height() - _zoom_indicator.height();
		painter.drawImage( x,y, _zoom_indicator );
	}
}

void ImageArea::resizeEvent(QResizeEvent *event)
{
	if ( _posx == 0 && _posy == 0 )
	{
		_previous_zoom = _computeFitZoom( _previous );
		_current_zoom = _computeFitZoomWithRotation( _current );
		_next_zoom = _computeFitZoom( _next );
	}
	update();
	QWidget::resizeEvent(event);
}

bool ImageArea::event( QEvent * event )
{
	if ( g_config.multitouch )
		return event_multitouch( event );
	return event_singletouch( event );
}

bool ImageArea::event_multitouch( QEvent * event )
{
	switch (event->type())
	{
		case QEvent::TouchBegin:
		case QEvent::TouchUpdate:
		case QEvent::TouchEnd:
		{
			QTouchEvent *touchEvent = static_cast<QTouchEvent *>(event);
			QList<QTouchEvent::TouchPoint> touchPoints = touchEvent->touchPoints();

			if ( event->type() == QEvent::TouchBegin && touchPoints.count() == 1 )
			{
				_resetTouchParams();
			}

			if ( touchPoints.count() == 1
			     && !_allow_zoom && !_allow_pan
			     && !_changing )
			{
				const QTouchEvent::TouchPoint &touchPoint0 = touchPoints.first();
				onPan( touchPoint0, event->type() == QEvent::TouchEnd );
			}

			if ( touchPoints.count() >= 2 && !_allow_drag && !_changing)
			{
				_two_fingers = true;

				const QTouchEvent::TouchPoint &touchPoint0 = touchPoints.first();
				const QTouchEvent::TouchPoint &touchPoint1 = touchPoints.last();
				onTwoFingers( touchPoint0, touchPoint1 );

				_limitPan();
				update();
			}

			if ( event->type() == QEvent::TouchEnd )
			{
				if ( !_allow_drag && !_allow_pan && !_allow_zoom
					&& touchPoints.count() >= 1 )
				{
					const QTouchEvent::TouchPoint &touchPoint0 = touchPoints.first();
					int x = (int)touchPoint0.pos().x();
					int y = (int)touchPoint0.pos().y();
					onTap(x,y);
					update();
				}
				if ( touchPoints.count() == 1 ) _allow_drag = false;
			}
		}
		break;

		case QEvent::MouseButtonPress:
		{
			QMouseEvent * mouseEvent = static_cast<QMouseEvent *>(event);
			if ( mouseEvent->button() == Qt::RightButton )
			{
				int x = mouseEvent->x();
				int y = mouseEvent->y();
				onTap(x,y);
				update();
			}
		}
		break;

		case QEvent::MouseButtonDblClick:
			if ( g_config.double_click_fullscreen )
				emit changeFullscreen();
			break;

		default:
			return QWidget::event(event);
	}
	return true;
}

void ImageArea::onPan( const QTouchEvent::TouchPoint & point, bool end )
{
	int dx = point.pos().x() - point.startPos().x();
	if ( !_two_fingers && abs(dx) > g_config.drag_sensitivity )
	{
		_allow_drag = true;
		_posx = _posy = 0;
		_rotation = 0.0;
		_current_zoom = _computeFitZoom( _current );
		_previous_zoom = _computeFitZoom( _previous );
		_next_zoom = _computeFitZoom( _next );
		_timer.start( 30 );
	}

	if ( _allow_drag )
	{
		_drag_offset = _initial_drag_offset + dx;
						
		update();
		if ( end )
		{
			if ( dx < -g_config.flip_distance && _current_index < _files.size() - 1
			     && _load_thread.isIdle() )
			{
				_drag_offset = dx + width();
				_moveForward();
				_changing = true;
			}
			if ( dx > g_config.flip_distance && _current_index > 0
			     && _load_thread.isIdle() )
			{
				_drag_offset = dx - width();
				_moveBack();
				_changing = true;
			}
		}
	}

}

void  ImageArea::onTwoFingers( const QTouchEvent::TouchPoint tp0,
	const QTouchEvent::TouchPoint tp1 )
{
	qreal d_current = QLineF(tp0.pos(), tp1.pos()).length();
	qreal d_start = QLineF(tp0.startPos(), tp1.startPos()).length();

	qreal alpha = QLineF( tp1.startPos(), tp0.startPos() ).angle()
		- QLineF( tp1.pos(), tp0.pos() ).angle();

	// decide operation type
	if ( !_allow_pan &&
	     ( fabs(d_current-d_start) > 2*g_config.drag_sensitivity || fabs(alpha) > 10.0 ) )
	{
		_allow_zoom = true;
	}

	#define NORM_ANGLE(r) ( (r) >= 360.0 ?\
		((r) - 360.0) : ( (r) < 0 ? ((r) + 360.0) : (r) ) );
	#define RAD2DEG(r) ( (r)* 180.0 / M_PI )
	#define DEG2RAD(r) ( (r)* M_PI / 180.0 )

	if ( _allow_zoom )
	{
		// compute xa,ya,xb,yb
		/*qreal cx = (qreal)( width() / 2 );
		qreal cy = (qreal)( height() / 2 );
		double roti = DEG2RAD(_initial_rotation);
		qreal ci = (qreal)cos( roti );
		qreal si = (qreal)sin( roti );
		qreal r0x = ( tp0.startPos().x() - _initial_posx - cx ) / _initial_zoom;
		qreal r1x = ( tp1.startPos().x() - _initial_posx - cx ) / _initial_zoom;
		qreal r0y = ( tp0.startPos().y() - _initial_posy - cy ) / _initial_zoom;
		qreal r1y = ( tp1.startPos().y() - _initial_posy - cy ) / _initial_zoom;
		qreal xa =   r0x * ci + r0y * si;
		qreal ya = - r0x * si + r0y * ci;
		qreal xb =   r1x * ci + r1y * si;
		qreal yb = - r1x * si + r1y * ci;

		qreal dx = (qreal)_current->width() /2 ;
		qreal dy = (qreal)_current->height() / 2;
		printf("%lf,%lf - %lf,%lf\n", xa+dx,ya+dy,xb+dx,yb+dy);

		// compute the rotation through other means
		//_rotation = _initial_rotation - alpha;
		//if ( _rotation > 360.0 ) _rotation -= 360.0;
		//if ( _rotation < 0 ) _rotation += 360.0;
		//double rot = _rotation * M_PI / 180.0;
		//qreal c = (qreal)cos( rot );
		//qreal s = (qreal)sin( rot );

		qreal dxp = xa - xb;
		qreal dyp = ya - yb;
		qreal dtx = tp0.pos().x() - tp1.pos().x();
		qreal dty = tp0.pos().y() - tp1.pos().y();
		qreal ks = dyp * dty - dxp * dtx;
		qreal kc = dyp * dtx - dxp * dty;
		qreal c_i = ks / sqrt( ks*ks + kc*kc );
		qreal rot1 = acos(c_i);
		qreal rot2 = acos(-c_i);
		qreal rot1deg = RAD2DEG(rot1);
		qreal rot2deg = RAD2DEG(rot2);
		rot1deg = NORM_ANGLE( fabs(rot1deg - _rotation) );
		rot2deg = NORM_ANGLE( fabs(rot2deg - _rotation) );
		qreal rot = ( rot1deg < rot2deg ? rot1 : rot2 );
		qreal c = (qreal)cos( rot );
		qreal s = (qreal)sin( rot );
		_rotation = RAD2DEG(rot);
		_rotation = NORM_ANGLE(_rotation);

		// compute _current_zoom,_posx,_posy
		qreal h0x = xa * c - ya * s;
		qreal h0y = xa * s + ya * c;
		qreal h1x = xb * c - yb * s;
		qreal h1y = xb * s + yb * c;
		qreal t0x = tp0.pos().x() - cx;
		qreal t0y = tp0.pos().y() - cy;
		qreal t1x = tp1.pos().x() - cx;
		qreal t1y = tp1.pos().y() - cy;

		qreal z = 0;
		qreal difx = h0x - h1x;
		qreal dify = h0y - h1y;
		if ( fabs(difx) > fabs(dify) )
			z = ( t0x - t1x ) / difx;
		else
			z = ( t0y - t1y ) / dify;
		qreal px = t0x - z * h0x;
		qreal py = t0y - z * h0y;

		// copy parameters
		_current_zoom = (double)z;
		_posx = (int)px;
		_posy = (int)py;
		//==> _current_zoom,_rotation,_posx,_posy
		*/

		_current_zoom = _initial_zoom * d_current / d_start;
		_limitZoom(_current_zoom);

		qreal r = - alpha;

		qreal xp_i = ( tp0.startPos().x() + tp1.startPos().x() ) / 2;
		qreal yp_i = ( tp0.startPos().y() + tp1.startPos().y() ) / 2;
		qreal xp_c = ( tp0.pos().x() + tp1.pos().x() ) / 2;
		qreal yp_c = ( tp0.pos().y() + tp1.pos().y() ) / 2;
		double rot = - _initial_rotation * M_PI / 180.0;
		double rot1 = r * M_PI / 180.0;
		qreal cx = (qreal)( width() / 2 );
		qreal cy = (qreal)( height() / 2 );
		qreal rx = ( xp_i - _initial_posx - cx ) / _initial_zoom ;
		qreal ry = ( yp_i - _initial_posy - cy ) / _initial_zoom ;
		qreal c = (qreal)cos( rot );
		qreal s = (qreal)sin( rot );
		qreal c1 = (qreal)cos( rot1 );
		qreal s1 = (qreal)sin( rot1 );
		qreal xa =   rx * c + ry * s;
		qreal ya = - rx * s + ry * c;
		qreal px = xp_c - _current_zoom * ( xa * c1 - ya * s1 ) - cx;
		qreal py = yp_c - _current_zoom * ( xa * s1 + ya * c1 ) - cy;

		_posx = (int)px;
		_posy = (int)py;
		if ( g_config.allow_rotation )
		{
			_rotation = _initial_rotation + r;
			if ( _rotation > 360.0 ) _rotation -= 360.0;
			if ( _rotation < 0 ) _rotation += 360.0;
		}

	} else {
		int dx0 = tp0.pos().x() - tp0.startPos().x();
		int dy0 = tp0.pos().y() - tp0.startPos().y();
		if ( dx0 > g_config.drag_sensitivity || dy0 > g_config.drag_sensitivity )
			_allow_pan = true;
		_posx = _initial_posx + dx0;
		_posy = _initial_posy + dy0;
	}
}

bool ImageArea::event_singletouch( QEvent * event )
{
	switch (event->type())
	{
		case QEvent::MouseButtonPress:
		{
			_initial_posx = _posx;
			_initial_posy = _posy;
			_initial_drag_offset = _drag_offset;
			_allow_drag = false;
			_allow_zoom = false;
			_allow_pan = false;

			QMouseEvent * mouseEvent = static_cast<QMouseEvent *>(event);
			//if ( mouseEvent->button() == Qt::LeftButton )
			{
				_mouse_drag = true;
				_mouse_start_x = mouseEvent->x();
				_mouse_start_y = mouseEvent->y();
			}
			update();
			break;
		}

		case QEvent::MouseButtonRelease:
		{
			QMouseEvent * mouseEvent = static_cast<QMouseEvent *>(event);
			//if ( mouseEvent->button() != Qt::LeftButton )
			//	break;
			if ( _allow_drag )
			{
				int dx = mouseEvent->x() - _mouse_start_x;

				if ( dx < -g_config.flip_distance && _current_index < _files.size() - 1
				    && !_changing && _load_thread.isIdle() )
				{
					_drag_offset = dx + width();
					_moveForward();
					_changing = true;
				
				} else if ( dx > g_config.flip_distance && _current_index > 0
				     && !_changing && _load_thread.isIdle() )
				{
					_drag_offset = dx - width();
					_moveBack();
					_changing = true;
				
				} else {
					_drag_offset = dx;
					_changing = true;
					_timer.start(30);
				}
			} else if ( !_allow_zoom && !_allow_pan ) {
				int x = mouseEvent->x();
				int y = mouseEvent->y();
				onTap(x,y);
			}

			_mouse_drag = _allow_pan = _allow_zoom = false;
			update();
			break;
		}

		case QEvent::MouseMove:
		{
			QMouseEvent * mouseEvent = static_cast<QMouseEvent *>(event);

			if ( !_changing && _mouse_drag )
			{
				int zby1 = ( height() - _zoom_in_icon.height() );
				int zby2 = height();
				bool zoom_zone = ( !g_config.multitouch
					&& _show_ui
					&& mouseEvent->y() > zby1
					&& mouseEvent->y() < zby2 );
				_allow_zoom = _allow_zoom || ( zoom_zone && !_allow_drag );
				
				double fit_zoom = _computeFitZoomWithRotation(_current);
				double min_zoom = ( fit_zoom < 1.0 ? fit_zoom : 1.0 );
				double max_zoom = g_config.max_zoom;
				
				if ( _allow_zoom )
				{
					// zoom
					
					int posx = mouseEvent->x() - _zoom_out_icon.width();
					int maxx = width() - _zoom_in_icon.width() - _zoom_out_icon.width();
					if ( posx < 0 ) posx = 0;
					if ( posx > maxx ) posx = maxx;
					
					_current_zoom = min_zoom + (double)posx * (max_zoom - min_zoom) / (double) maxx;
					_limitPan();
					update();
					
				} else if ( _current_zoom != min_zoom ) {
				
					// pan
					_allow_pan = true;
					int dx0 = mouseEvent->x() - _mouse_start_x;
					int dy0 = mouseEvent->y() - _mouse_start_y;
					if ( dx0 > g_config.drag_sensitivity || dy0 > g_config.drag_sensitivity )
						_allow_pan = true;
					if ( _allow_pan )
					{
						_posx = _initial_posx + dx0;
						_posy = _initial_posy + dy0;
						_limitPan();
					}
				} else {
					
					// drag to next image
				
					int dx = mouseEvent->x() - _mouse_start_x;
					if ( abs(dx) > g_config.drag_sensitivity )
					{
						_allow_drag = true;
						_current_zoom = _computeFitZoom( _current );
						_posx = _posy = 0;
						_rotation = 0.0;
					}
					if ( _allow_drag )
					{
						_drag_offset = _initial_drag_offset + dx;
						update();
					}
				}
			}

			break;
		}

		case QEvent::MouseButtonDblClick:
			if ( g_config.double_click_fullscreen )
				emit changeFullscreen();
			break;

		default:
			return QWidget::event(event);
	}
	return true;
}

void ImageArea::onTap( int x, int y )
{
	TouchUIAction action = TOUCH_ACTION_NONE;
	if ( _show_ui )
		action = _ui.getAction(x,y);

	_handleTouchAction( action );

	if ( action == TOUCH_ACTION_NONE )
		_show_ui = !_show_ui;
}

void ImageArea::onTimer( void )
{
	if ( !_allow_drag || _changing )
	{
		if ( _drag_offset > 0 )
		{
			_drag_offset -= g_config.dragspeed; //50;
			if ( _drag_offset <= 0 )
			{
				_drag_offset = 0;
			}
		} else if ( _drag_offset < 0 ) {
			_drag_offset += g_config.dragspeed; //50;
			if ( _drag_offset > 0 )
			{
				_drag_offset = 0;
			}
		}
		update();
	}
	if ( _drag_offset == 0 ) _changing = false;

	//if ( _drag_offset == 0 && _load_thread.isIdle() )
	//	_timer.stop();
}

/*******************************************************************************
* PRIVATE METHODS
*******************************************************************************/

void  ImageArea::_loadImage( int k, QImage ** dest, double * zoom_dest )
{
	*dest = NULL;
	*zoom_dest = 1.0;
	if ( k < 0 || k >= _files.size() )
		return;

	ImageLoadItem ili;
	ili.name = _files[k];
	ili.destination = dest;
	ili.zoom = zoom_dest;
	ili.w = width();
	ili.h = height();
	_load_thread.addLoadImage(ili);
}

double ImageArea::_computeFitZoom( QImage * img )
{
	if ( img == NULL ) return 1.0f;
	double xz = (double)width() / (double)img->width();
	double yz = (double)height() / (double)img->height();
	double z = ( xz < yz ? xz : yz );
	return z > 1.0f ? 1.0f : z;
}

double ImageArea::_computeFitZoomWithRotation( QImage * img )
{
	if ( img == NULL ) return 1.0f;
	double rot = _rotation * M_PI / 180.0;
	qreal c = (qreal)cos( rot );
	qreal s = (qreal)sin( rot );
	double w = (double)img->width();
	double h = (double)img->height();
	double img_w = fabs(w * c + h * s);
	double img_h = fabs(w * s + h * c);
	double xz = (double)width() / img_w;
	double yz = (double)height() / img_h;
	double z = ( xz < yz ? xz : yz );
	return z > 1.0f ? 1.0f : z;
}

void ImageArea::_handleTouchAction( TouchUIAction action )
{
	switch ( action )
	{
		case TOUCH_ACTION_EXIT:
			emit closeOnTouch();
			break;
		case TOUCH_ACTION_FULLSCREEN:
			emit changeFullscreen();
			break;
		case TOUCH_ACTION_ZOOM_IN:
			zoomIn();
			break;
		case TOUCH_ACTION_ZOOM_OUT:
			zoomOut();
			break;
		case TOUCH_ACTION_ZOOM_FIT:
			zoomToFit();
			break;
		case TOUCH_ACTION_ZOOM_ORIGINAL:
			zoomToOriginalSize();
			break;
		case TOUCH_ACTION_ROTATE_LEFT:
			rotateLeft();
			break;
		case TOUCH_ACTION_ROTATE_RIGHT:
			rotateRight();
			break;
		case TOUCH_ACTION_GO_FIRST:
			first();
			break;
		case TOUCH_ACTION_GO_LAST:
			last();
			break;
		case TOUCH_ACTION_GO_NEXT:
			forward();
			break;
		case TOUCH_ACTION_GO_PREVIOUS:
			back();
			break;
		case TOUCH_ACTION_OPEN:
			emit loadFile();
			break;
		default:
			break;
	}
}

void ImageArea::_limitZoom( double & zoom )
{
	if ( zoom > g_config.max_zoom )
		zoom = g_config.max_zoom;
	else if ( zoom < g_config.min_zoom )
		zoom = g_config.min_zoom;
}

void ImageArea::_limitPan( void )
{
	if ( _current == NULL ) return;
	//int sw = (int)( (double)_current->width() * _current_zoom );
	//int sh = (int)( (double)_current->height() * _current_zoom );
	
	double rot = _rotation * M_PI / 180.0;
	qreal c = (qreal)cos( rot );
	qreal s = (qreal)sin( rot );
	double w = (double)_current->width() * _current_zoom;
	double h = (double)_current->height() * _current_zoom;
	int sw = (int)fabs(w * c + h * s);
	int sh = (int)fabs(w * s + h * c);
	
	if ( sw > width() )
	{
		int max_x = ( sw - width() ) / 2;
		int min_x = - sw + max_x + width();
		if ( _posx > max_x ) _posx = max_x;
		if ( _posx < min_x ) _posx = min_x;
	} else {
		_posx = 0;
	}
	if ( sh > height() )
	{
		int max_y = ( sh - height() ) / 2;
		int min_y = - sh + max_y + height();
		if ( _posy > max_y ) _posy = max_y;
		if ( _posy < min_y ) _posy = min_y;
	} else {
		_posy = 0;
	}
}

void ImageArea::_resetTouchParams( void )
{
	_initial_posx = _posx;
	_initial_posy = _posy;
	_initial_drag_offset = _drag_offset;
	_initial_zoom = _current_zoom;
	_initial_rotation = _rotation;
	_allow_zoom = false;
	_allow_pan = false;
	_allow_drag = false;
	_two_fingers = false;
}

void ImageArea::_reloadAll( void )
{
	delete _previous;
	delete _current;
	delete _next;
	_previous = _current = _next = NULL;

	_posx = _posy = 0;
	_rotation = 0.0;

	_loadImage( _current_index-1, &_previous, &_previous_zoom );
	_loadImage( _current_index, &_current, &_current_zoom );
	_loadImage( _current_index+1, &_next, &_next_zoom );
}

void ImageArea::_moveForward( void )
{
	delete _previous;
	_current_index++;
	_posx = _posy = 0;
	_rotation = 0.0;

	_previous = _current;
	_previous_zoom = _current_zoom;

	_current = _next;
	_current_zoom = _next_zoom;

	_loadImage( _current_index+1, &_next, &_next_zoom );
}

void ImageArea::_moveBack( void )
{
	delete _next;
	_current_index--;
	_posx = _posy = 0;
	_rotation = 0.0;

	_next = _current;
	_next_zoom = _current_zoom;

	_current = _previous;
	_current_zoom = _previous_zoom;

	_loadImage( _current_index-1, &_previous, &_previous_zoom );
}

QTransform ImageArea::_getCurrentTransform( void )
{
	QTransform tr;

	qreal dx = (qreal)_current->width();
	qreal dy = (qreal)_current->height();

	tr.translate( (qreal)width() / 2, (qreal)height() / 2 );
	tr.translate( (qreal)(_posx + _drag_offset), (qreal)(_posy) );
	tr.rotate(-_rotation);
	tr.scale( _current_zoom, _current_zoom );
	tr.translate( -dx/2,-dy/2);

	return tr;
}
